__author__ = 'bkeroack'

class InvalidKeyPairType(Exception):
    pass
class InvalidPublicKey(Exception):
    pass
class InvalidPrivateKey(Exception):
    pass

class SaltServerNotAccessible(Exception):
    pass

